/****************************************************************************/
/*                                                                          */
/*  Function:  Show how to define and use a simple C++ class -- Time        */
/*                                                                          */
/*  Author:   Bary W Pollack                                                */
/*                                                                          */
/*  File:     TimeClas.cpp                                                  */
/*                                                                          */
/*  Date:     07/21/98      Revised:    10/19/01                            */
/*                                                                          */
/****************************************************************************/

#include <iostream>                         // for standard input/output
#include <iomanip>                          // for setw()
#include <cctype>                           // for character-testing routines
#include <cstdlib>                          // for abort()

using namespace std;

// Function prototypes
void Error (const char *sName, const char *sMessage);
int  Digit2Int (const char c);
void ReadHour (istream& ins, int& nHour);
void ReadMinute (istream& ins, int& nMinute);

/****************************************************************************/

class Time {    // NOTE: Layout exhibits POOR style; I wanted it all to fit...

  public:

    Time (void)                             // default constructor
    {  nHour = nMinute = 0;  }              // initializes to 00:00 (midnight)

    Time (const int nHr, const int nMin)    // constructor (hh:mm)
    {  nHour = nHr;  nMinute = nMin;  }     // initializes to nHr:nMin

    int GetHour (void) const                // access function:  get hour
    {  return nHour;  }

    int GetMinute (void) const              // access function:  get minute
    {  return nMinute;  }

    void SetHour (const int nHr)            // access function:  set hour
    {  nHour = nHr;  }

    void SetMinute (const int nMin)         // access function:  set minute
    {  nMinute = nMin;  }

    void SetTime (const int nHr, const int nMin)  // access:  set hh:mm
    {  nHour = nHr;  nMinute = nMin;  }

    void ShowTime (void) const              // display time as h:m
    {  cout << "nHour=" << nHour << " nMinute=" << nMinute << endl;  }

    void Advance (const int nMin)           // advances by nMin minutes
    {
        int nTotalMin = nMinute + nMin;
        nMinute = nTotalMin % 60;
        int nAdjustHour = nTotalMin / 60;
        nHour = (nHour + nAdjustHour) % 24;
    }

    void Advance (const int nHr, const int nMin)    // advances by nHr, nMin
    {
        nHour = (nHour + nHr) % 24;
        Advance (nMin);
    }

    friend ostream& operator << (ostream& outs, const Time& time)
    {
        outs << setw(2) << time.nHour << ':';
        if (time.nMinute < 10)
            outs << '0';
        outs << time.nMinute;
        return outs;
    }

    friend istream& operator >> (istream& ins, Time& time)
    {
        ReadHour (ins, time.nHour);
        ReadMinute (ins, time.nMinute);
        return ins;
    }

    friend bool operator == (const Time& time1, const Time& time2)
    {  return (time1.nHour == time2.nHour && time1.nMinute == time2.nMinute);  }

    friend bool operator != (const Time& time1, const Time& time2)
    {  return (time1.nHour != time2.nHour || time1.nMinute != time2.nMinute);  }

  private:

    int nHour;                              // the hour
    int nMinute;                            // the minute

};  // end class Time

/****************************************************************************/
/*    Error -- issues an error message, then aborts the program             */
/****************************************************************************/

void Error (const char *sName, const char *sMessage)
{
    cerr << endl << sName << " Error: " << sMessage << endl << endl;
    abort ();
}

/****************************************************************************/
/*    Digit2Int -- converts a digit character into its numeric equivalent   */
/****************************************************************************/

int Digit2Int (const char c)
{
    if (!isdigit (c))
        Error ("Digit2Int", "encountered a non-digit");
    return (int (c) - int ('0'));
}

/****************************************************************************/
/*    ReadHour -- reads the hh portion of hh:mm, including the :            */
/****************************************************************************/

void ReadHour (istream& ins, int& nHour)
{
    char c1, c2;

    while (ins >> c1, isspace (c1))
        ;
    ins >> c2;
    if (! (isdigit (c1) && isdigit (c2) || c2 == ':'))
        Error ("ReadHour", "Improper format: expecting hh:mm");

    if (c2 == ':')
        nHour = Digit2Int (c1);
    else
    {
        nHour = 10 * Digit2Int (c1) + Digit2Int (c2);
        ins >> c2;
        if (c2 != ':')
            Error ("ReadHour", "Improper format: expecting hh:mm; no ':'");
    }

    if (nHour < 0 || nHour > 23)
        Error ("ReadHour", "Hour out of range -- it must be in 0..23");
}

/****************************************************************************/
/*    ReadMinute -- reads the mm portion of hh:mm, assumes ReadHour() called*/
/****************************************************************************/

void ReadMinute (istream& ins, int& nMinute)
{
    char c1, c2;

    ins >> c1 >> c2;
    if (!isdigit (c1) || !isdigit (c2))
        Error ("ReadMinute", "Improper format: expecting hh:mm");

    nMinute = 10 * Digit2Int (c1) + Digit2Int (c2);
    if (nMinute < 0 || nMinute > 59)
        Error ("ReadMinute", "Minute out of range -- it must be in 0..59");
}

/****************************************************************************/
/*    main -- show how the various functions are used...                    */
/****************************************************************************/

int main (void)
{
    cout << "Demonstration of a Time class..." << endl << endl;

    Time time1, time2 (10, 2);
    time1.ShowTime ();
    time2.ShowTime ();
    time1 = Time (9, 57);

    cout << endl << "The clock is ticking..." << endl << endl;
    while (! (time1 == time2))
    {
        time1.Advance (1);
        cout << "Now it's " << time1 << " ..." << endl;
    }
    cout << endl << "Finally " << time2 << "!" << endl
         << endl << "Now, let's move closer to midnight..." << endl << endl;

    time1.Advance (13, 54);
    time2.SetTime (0, 1);

    while (time1 != time2)
    {
        time1.Advance (1);
        cout << "Now it's " << time1 << " ..." << endl;
    }

    cout << endl << "Tell me, what time is it? ";
    cin >> time1;
    time2 = time1;
    time2.Advance (1, 15);
    cout << "Thanks!  1 hour, 15 minutes later it'll be " << time2 << endl;
    time2.Advance (12, 46);
    cout << "And, 12 hours, 46 minutes later it'll be   " << time2 << endl;

    cout << endl << "That's all, folks!" << endl << endl;

    return 0;
}

/****************************************************************************/

/*.....
Demonstration of a Time class...

nHour=0 nMinute=0
nHour=10 nMinute=2

The clock is ticking...

Now it's  9:58 ...
Now it's  9:59 ...
Now it's 10:00 ...
Now it's 10:01 ...
Now it's 10:02 ...

Finally 10:02!

Now, let's move closer to midnight...

Now it's 23:57 ...
Now it's 23:58 ...
Now it's 23:59 ...
Now it's  0:00 ...
Now it's  0:01 ...

Tell me, what time is it? 12:34
Thanks!  1 hour, 15 minutes later it'll be 13:49
And, 12 hours, 46 minutes later it'll be    2:35

That's all, folks!

..... */

